(* A theory about parse trees *)
open HolKernel boolLib bossLib Parse
open stringTheory;

val _ = new_theory "parseTree";

(* is non terminal a single char string?? *)
val _ = Hol_datatype `nonTerminal = NT of string`;

(* what if terminal is also of type string or 'a?? *)
val _ = Hol_datatype `terminal = TM of string`;

val tmnlToStr = Define `(tmnlToStr (TM s) = s)`;

val nonTmnlToStr = Define `(nonTmnlToStr (NT s) = s)`;

val _ = Hol_datatype `ptree = Leaf of terminal | Node of nonTerminal => ptree list`;

val ptsize_def = Define`  (ptsize (Leaf tmnl) = 1) 
/\  (ptsize (Node nt ptlist) = 1 + ptsizel ptlist) 
/\  (ptsizel [] = 0) 
/\     (ptsizel (h::t) = ptsize h + ptsizel t) `
val _ = BasicProvers.export_rewrites ["ptsize_def"]

val ptsize_better = prove(
  ``ptsize (Node nt ptlist) = 1 + SUM (MAP ptsize ptlist)``,
  SRW_TAC [][] THEN Induct_on `ptlist` THEN
  SRW_TAC [][]);


val size_nonzero = prove(
  ``!pt. ptsize pt > 0``,
  Induct THENL [
    SRW_TAC [][ptsize_def],
    SRW_TAC [numSimps.ARITH_ss][ptsize_def]
  ]);

(* <<HOL warning: Type.mk_type: more than one possibility>>
<<HOL message: Defined type: "option">> *)

(* fun parseTreeToRule :: ptree => rule *)

(* ptreeToRule :: ptree => [symbol, [symbol]] *)

(* try checking the rules while recursing over the tree + using sets
in HOL rather than converting to list *)

val getSymbols = Define `(getSymbols [] = []) /\ (getSymbols [Leaf tmnl] = [tmnlToStr tmnl]) /\ (getSymbols [Node nt ptlist] = [nonTmnlToStr nt]) /\ (getSymbols ((Leaf tmnl)::t) = (tmnlToStr tmnl) :: (getSymbols t)) /\ (getSymbols ((Node nt ptlist)::t) = (nonTmnlToStr nt) :: (getSymbols t))`;

val ptreeToRules = Define `(ptreeToRules (Leaf tmnl) = [(tmnlToStr tmnl, [])]) /\ (ptreeToRules (Node nt ptlist) = (nonTmnlToStr nt, getSymbols ptlist) :: (ptreeToRules2 ptlist)) /\ (ptreeToRules2 [] = []) /\ (ptreeToRules2 (h::t) = (ptreeToRules h) ++ (ptreeToRules2 t))`;

val exists = Define `(exists elem [] = F) /\ (exists elem (h::t) = ((elem = h)) \/ (exists elem t))`;

(* This should be checking against an actual rule format rather than a string format!!!! *)
val checkRules = Define `(checkRules [] g = T) /\ (checkRules (h::t) g = (exists h g) /\ checkRules t g)`;

val validptree = Define `(validptree pt g = (checkRules (ptreeToRules pt) g))`;

val leaves_def = Define `(leaves (Leaf tmnl) = [tmnlToStr (tmnl)]) /\ (leaves (Node nt ptlist) = (nonTmnlToStr nt) :: cleaves ptlist) /\ (cleaves [] = []) /\ (cleaves (h::t) = leaves h ++ cleaves t)`;

(* write prettyPrinter for parse tree *)


(* Checking the validity of the parse tree while recursing over it *)

val validptreerec = Define `(validptreerec (Leaf tmnl) rset = (tmnlToStr tmnl, []) IN rset) /\ (validptreerec (Node nt ptlist) rset = (nonTmnlToStr nt, getSymbols ptlist) IN rset /\ validptreerec2 ptlist rset) /\ (validptreerec2 [] rset = T) /\ (validptreerec2 (h::t) rset = validptreerec h rset /\ validptreerec2 t rset)`;

val yield_def = Define `yield ptree = FOLDR STRCAT "" (leaves ptree)`;

val _ = export_theory ();

